/*
 * gedit-code-listing-plugin-panel.c
 *
 * Copyright (C) 2005 - Jesse van den Kieboom 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
 
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>

#include "gedit-code-listing-plugin-panel.h"
#include "gedit-code-listing-plugin-parser.h"

#include <gedit/gedit-utils.h>
#include <gedit/gedit-commands.h>
#include <gedit/gedit-file-selector-util.h>
#include <gedit/gedit-debug.h>
#include <gedit/gedit-prefs-manager.h>
#include <gedit/gedit-prefs-manager-private.h>
#include <gconf/gconf.h>

#include <gdk/gdkkeysyms.h>
#include <glib/gi18n.h>

#define GEDIT_CODE_LISTING_PLUGIN_PANEL_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), GEDIT_TYPE_CODE_LISTING_PLUGIN_PANEL, GeditCodeListingPluginPanelPrivate))
#define GPM_PLUGIN_CODE_LISTING GPM_PLUGINS_DIR "/code-listing"
#define COMBO_BOX_TAGS_TOOLTIP "Select a previous opened tag file for loading, or load a new tag file"
#define PANEL_GTK_TREE_MODEL(panel) gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(panel->priv->tree_store_full))
#define PANEL_GTK_TREE_MODEL_SEARCH(panel) gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(panel->priv->list_store_search))

enum TagHint
{
  TAG_HINT_CLASS = 0,
  TAG_HINT_STRUCT,
  TAG_HINT_ENUM,
  TAG_HINT_DEFINE,
  TAG_HINT_VARIABLE,
  TAG_HINT_TYPEDEF,
  TAG_HINT_FUNCTION,
  TAG_HINT_NAMESPACE,
  TAG_HINT_NONE,
  TAG_HINT_FIELD,
  TAG_HINT_PROTO,
  TAG_HINT_COUNT
};

static void gedit_code_listing_plugin_panel_gslist_free(GSList *list);
static void gedit_code_listing_plugin_panel_free_pixbufs(GeditCodeListingPluginPanel *panel);
static gboolean gedit_code_listing_plugin_panel_free_entry_parent(gpointer key, gpointer value, gpointer user_data);
static void gedit_code_listing_plugin_panel_free_entry_parents(GeditCodeListingPluginPanel *panel);
static void gedit_code_listing_plugin_panel_process_tag(GeditCodeListingPluginPanel *panel, tagEntry *entry, enum TagHint hint);

enum
{
  COLUMN_TAG_PIXBUF = 0,
	COLUMN_TAG_NAME,
	COLUMN_TAG_ENTRY,
	COLUMN_TAG_HINT,
	COLUMN_TAG_SORT,
  COLUMN_TAG_STYLE,
  COLUMN_TAG_WEIGHT,
	NUM_COLUMNS
};

enum
{
  SEARCH_COLUMN_TAG_PIXBUF = 0,
	SEARCH_COLUMN_TAG_NAME,
	SEARCH_COLUMN_TAG_ENTRY,
	SEARCH_NUM_COLUMNS
};

typedef struct _tagEntryProxy {
  GtkTreePath *path;
  enum TagHint hint;
} tagEntryProxy;

struct _GeditListingType {
  const enum TagHint hint;
  const gchar *name;
  const gchar *pixbuf;
};

#define PIXMAP_DIR "../plugins/codelisting/pixmaps/"

static const struct _GeditListingType GeditListingTypes[] = {
  {TAG_HINT_CLASS, "Classes", PIXMAP_DIR "class.png"},
  {TAG_HINT_STRUCT, "Structs", PIXMAP_DIR "struct.png"},
  {TAG_HINT_ENUM, "Enums", PIXMAP_DIR "enum.png"},
  {TAG_HINT_DEFINE, "Defines", PIXMAP_DIR "default.png"},
  {TAG_HINT_VARIABLE, "Variables", PIXMAP_DIR "variable.png"},
  {TAG_HINT_TYPEDEF, "Typedefs", PIXMAP_DIR "default.png"},
  {TAG_HINT_FUNCTION, "Functions", PIXMAP_DIR "method.png"},
  {TAG_HINT_NAMESPACE, "Namespaces", PIXMAP_DIR "namespace.png"},
  {TAG_HINT_NONE, NULL, PIXMAP_DIR "default.png"},
  {TAG_HINT_FIELD, NULL, PIXMAP_DIR "field.png"},
  {TAG_HINT_PROTO, NULL, PIXMAP_DIR "default.png"}
};

struct _GeditCodeListingPluginPanelPrivate
{
	GeditWindow  *window;

  char *prev_path;
  GtkWidget *entry_search;
  
  GtkTreeModel *tree_store_full;
  GtkTreeModel *list_store_search;
  GtkTreeViewColumn *column_full;
  GtkTreeViewColumn *column_search;
  
  GtkWidget *tree_view_tags;
  GtkWidget *combo_box_tags;
  GtkWidget *menu;
  
  struct {
    GtkWidget *table;
    GtkWidget *label_file;
    GtkWidget *label_line;
  } table_info;
  
  struct {
    GtkWidget *progress;
    GtkWidget *label;
    GtkWidget *hbox;
    GtkWidget *button_cancel;
  } hbox_progress;
  
  GtkWidget *vbox_filling;
  
  GSList *tag_files;
  GdkPixbuf *pixbufs[TAG_HINT_COUNT];

  GeditCodeListingPluginParser *parser;
	GtkTooltips *tooltips;
	
	GHashTable *entry_parents;
	GList *prototypes;
	gint idle_search;
	gint search_id;
};

static GObjectClass *parent_class = NULL;
static GType type = 0;

enum
{
	PROP_0,
	PROP_WINDOW,
}; 

static void
set_window (GeditCodeListingPluginPanel *panel,
	    GeditWindow             *window)
{
	g_return_if_fail (panel->priv->window == NULL);
	g_return_if_fail (GEDIT_IS_WINDOW (window));
	
	panel->priv->window = window;	 
}

static void
gedit_code_listing_plugin_panel_set_property (GObject      *object,
					 guint         prop_id,
					 const GValue *value,
					 GParamSpec   *pspec)
{
	GeditCodeListingPluginPanel *panel = GEDIT_CODE_LISTING_PLUGIN_PANEL (object);

	switch (prop_id)

	{
		case PROP_WINDOW:
			set_window (panel, g_value_get_object (value));
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
gedit_code_listing_plugin_panel_get_property (GObject    *object,
					 guint       prop_id,
					 GValue     *value,
					 GParamSpec *pspec)
{
	GeditCodeListingPluginPanel *panel = GEDIT_CODE_LISTING_PLUGIN_PANEL (object);

	switch (prop_id)
	{
		case PROP_WINDOW:
			g_value_set_object (value,
					    GEDIT_CODE_LISTING_PLUGIN_PANEL_GET_PRIVATE (panel)->window);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;			
	}
}

static void
gedit_code_listing_plugin_panel_finalize (GObject *object)
{
  GeditCodeListingPluginPanel *panel = GEDIT_CODE_LISTING_PLUGIN_PANEL(object);
  
  gconf_client_set_list(gedit_prefs_manager->gconf_client, 
      GPM_PLUGIN_CODE_LISTING "/tag_files", GCONF_VALUE_STRING, panel->priv->tag_files, NULL);
	gedit_code_listing_plugin_panel_gslist_free(panel->priv->tag_files);
  
  gedit_code_listing_plugin_panel_free_entry_parents(panel);
  g_list_free(panel->priv->prototypes);
  
	if (panel->priv->parser != NULL) {
    g_object_unref(panel->priv->parser);
  }
	
	g_object_unref(panel->priv->tree_store_full);
	g_object_unref(panel->priv->list_store_search);
	g_object_unref(panel->priv->column_full);
	g_object_unref(panel->priv->column_search);
  g_object_unref(panel->priv->hbox_progress.button_cancel);
  g_object_unref(panel->priv->hbox_progress.hbox);
  g_object_unref(panel->priv->table_info.table);
  
	gedit_code_listing_plugin_panel_free_pixbufs(panel);
	
	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void 
gedit_code_listing_plugin_panel_class_init (GeditCodeListingPluginPanelClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);
	
	object_class->finalize = gedit_code_listing_plugin_panel_finalize;
	object_class->get_property = gedit_code_listing_plugin_panel_get_property;
	object_class->set_property = gedit_code_listing_plugin_panel_set_property;
	
	g_object_class_install_property (object_class,
					 PROP_WINDOW,
					 g_param_spec_object ("window",
							 "Window",
							 "The GeditWindow this GeditCodeListingPluginPanel is associated with",
							 GEDIT_TYPE_WINDOW,
							 G_PARAM_READWRITE |
							 G_PARAM_CONSTRUCT_ONLY));	
							      
	g_type_class_add_private (object_class, sizeof(GeditCodeListingPluginPanelPrivate));
}

static void
gedit_code_listing_plugin_panel_free_pixbufs(GeditCodeListingPluginPanel *panel)
{
  int i;
  
  for (i = 0; i < TAG_HINT_COUNT; i++) {
    if (panel->priv->pixbufs[i] != NULL) {
      g_object_unref(panel->priv->pixbufs[i]);
    }
  }
}

static void
gedit_code_listing_plugin_panel_gslist_free(GSList *list)
{
  GSList *iter;
  for (iter = list; iter; iter = iter->next) {
    g_free(iter->data);
  }
  
  g_slist_free(list);
}

static enum TagHint
gedit_code_listing_plugin_panel_tag_hint(tagEntry *entry)
{
  if (entry->kind) {
    switch (entry->kind[0]) {
      case 'f':
        return TAG_HINT_FUNCTION;
        break;
      case 'm':
        return TAG_HINT_FIELD;
        break;
      case 'e':
        return TAG_HINT_ENUM;
        break;
      case 'c':
        return TAG_HINT_CLASS;
        break;
      case 's':
        return TAG_HINT_STRUCT;
        break;
      case 'g':
        return TAG_HINT_ENUM;
        break;
      case 'v':
        return TAG_HINT_VARIABLE;
        break;
      case 'd':
        return TAG_HINT_DEFINE;
        break;
      case 't':
        return TAG_HINT_TYPEDEF;
        break;
      case 'p':
        return TAG_HINT_PROTO;
        break;
      case 'n':
        return TAG_HINT_NAMESPACE;
        break;
    }
  }
  
  return TAG_HINT_NONE;
}

static gboolean
gedit_code_listing_plugin_panel_tree_get_parent(GtkTreeModel *model, enum TagHint hint, GtkTreeIter *parent)
{
  GtkTreePath *path;
  
  if (hint >= TAG_HINT_NONE) {
    return FALSE;
  }
  
  path = gtk_tree_path_new_from_indices(hint, -1);
  gtk_tree_model_get_iter(model, parent, path);
  gtk_tree_path_free(path);
  
  return TRUE;
}

static gboolean
gedit_code_listing_plugin_panel_tree_find_tag(GtkTreeModel *model, gchar *key, enum TagHint hint, GtkTreeIter *parent, GtkTreeIter *iter)
{  
  tagEntry *entry;
  enum TagHint chint;
  
  if (gtk_tree_model_iter_children(model, iter, parent)) {
    do {
      gtk_tree_model_get(model, iter, COLUMN_TAG_ENTRY, &entry, COLUMN_TAG_HINT, &chint, -1);
      if (strcmp(entry->name, key) == 0 && chint == hint) {
        return TRUE;
      }
    } while (gtk_tree_model_iter_next(model, iter));
  }
  
  return FALSE;
}

static gboolean
gedit_code_listing_plugin_panel_tree_find_parent(GeditCodeListingPluginPanel *panel, GtkTreeModel *model, gchar *key, enum TagHint hint, GtkTreeIter *iter) 
{  
  tagEntryProxy *entry;
  GList *list;
  
  for (list = g_hash_table_lookup(panel->priv->entry_parents, key); list; list = list->next) {
    entry = (tagEntryProxy *)(list->data);
    
    if (entry->hint == hint) {
      gtk_tree_model_get_iter(model, iter, entry->path);
      return TRUE;
    }
  }
  
  return FALSE;
}

static void
gedit_code_listing_plugin_panel_tree_add_parent(GeditCodeListingPluginPanel *panel, gchar *key, enum TagHint hint, GtkTreeIter *iter) 
{
  tagEntryProxy *entry = g_new0(tagEntryProxy, 1);
  
  GList *list = g_hash_table_lookup(panel->priv->entry_parents, key);
  entry->path = gtk_tree_model_get_path(PANEL_GTK_TREE_MODEL(panel), iter);
  entry->hint = hint;
  g_hash_table_insert(panel->priv->entry_parents, key, g_list_prepend(list, entry));
}

static void
gedit_code_listing_plugin_panel_initialize_tree(GeditCodeListingPluginPanel *panel)
{
  int i;
  GtkTreeIter iter;
  gchar *sort;
  GtkTreeModelSort *tsort = GTK_TREE_MODEL_SORT(gtk_tree_view_get_model(GTK_TREE_VIEW(panel->priv->tree_view_tags)));
  GtkTreeStore *model = GTK_TREE_STORE(gtk_tree_model_sort_get_model(tsort));
  
  gtk_tree_store_clear(model);
  
  for (i = 0; i < TAG_HINT_NONE; i++) {
    sort = g_strdup_printf("%d", i);

    gtk_tree_store_append(model, &iter, NULL);
    gtk_tree_store_set(model, &iter, COLUMN_TAG_PIXBUF, panel->priv->pixbufs[i],
      COLUMN_TAG_NAME, GeditListingTypes[i].name, COLUMN_TAG_SORT, sort, COLUMN_TAG_WEIGHT, PANGO_WEIGHT_BOLD, -1);
    g_free(sort);
  }
}

static void
gedit_code_listing_plugin_panel_set_tag(GeditCodeListingPluginPanel *panel, GtkTreeStore *tstore, GtkTreeIter *iter, 
    tagEntry *entry, gchar *name, enum TagHint hint, gchar *sort) {
  
  if (hint >= 0 && hint < TAG_HINT_COUNT) {
    gtk_tree_store_set(tstore, iter, COLUMN_TAG_NAME, name, COLUMN_TAG_ENTRY, entry, COLUMN_TAG_HINT, hint, COLUMN_TAG_PIXBUF, panel->priv->pixbufs[hint], COLUMN_TAG_SORT, sort, -1);
  } else {
    gtk_tree_store_set(tstore, iter, COLUMN_TAG_NAME, name, COLUMN_TAG_ENTRY, entry, COLUMN_TAG_HINT, hint, COLUMN_TAG_SORT, sort, -1);
  }
}

static void
gedit_code_listing_plugin_panel_search_done(GeditCodeListingPluginParser *parser,
  GeditCodeListingPluginPanel *panel) {
  
  gtk_widget_set_sensitive(panel->priv->entry_search, TRUE);
  
  if (gtk_widget_get_parent(panel->priv->hbox_progress.hbox) == panel->priv->vbox_filling) {
    gtk_container_remove(GTK_CONTAINER(panel->priv->vbox_filling), panel->priv->hbox_progress.button_cancel);
    gtk_container_remove(GTK_CONTAINER(panel->priv->vbox_filling), panel->priv->hbox_progress.hbox);
  }
  if (gtk_widget_get_parent(panel->priv->table_info.table) != panel->priv->vbox_filling) {
    gtk_box_pack_start(GTK_BOX(panel->priv->vbox_filling), panel->priv->table_info.table, FALSE, TRUE, 0);
  }  
}

static void
gedit_code_listing_plugin_panel_tag_found(GeditCodeListingPluginParser *parser,
  tagEntry *entry, GeditCodeListingPluginPanel *panel) {
  long size, position;
  GtkTreeModel *model;
  GtkTreeIter iter;
  enum TagHint hint;
  
  gedit_code_listing_plugin_parser_search_progress(parser, &position, &size);
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(panel->priv->hbox_progress.progress),
    (double)(position) / (double)(size));
    
  if (entry != NULL) {
    model = PANEL_GTK_TREE_MODEL_SEARCH(panel);
    gtk_list_store_prepend(GTK_LIST_STORE(model), &iter);
    
    hint = gedit_code_listing_plugin_panel_tag_hint(entry);
    
    if (hint >= 0 && hint < TAG_HINT_COUNT) {
      gtk_list_store_set(GTK_LIST_STORE(model), &iter, SEARCH_COLUMN_TAG_NAME, entry->name, SEARCH_COLUMN_TAG_ENTRY, entry, SEARCH_COLUMN_TAG_PIXBUF, panel->priv->pixbufs[hint], -1);
    } else {
      gtk_list_store_set(GTK_LIST_STORE(model), &iter, SEARCH_COLUMN_TAG_NAME, entry->name, SEARCH_COLUMN_TAG_ENTRY, entry, -1);
    }
  } 
}

static void
gedit_code_listing_plugin_panel_parse_done(GeditCodeListingPluginParser *parser,
  GeditCodeListingPluginPanel *panel) {
  
  GList *list;
  tagEntry *entry;
  enum TagHint hint;
  GtkTreeIter iter;
  GtkTreeModel *model = PANEL_GTK_TREE_MODEL(panel);
  gboolean next;
  
  if (gtk_widget_get_parent(panel->priv->hbox_progress.hbox) == panel->priv->vbox_filling) {
    gtk_container_remove(GTK_CONTAINER(panel->priv->vbox_filling), panel->priv->hbox_progress.button_cancel);
    gtk_container_remove(GTK_CONTAINER(panel->priv->vbox_filling), panel->priv->hbox_progress.hbox);
  }
  if (gtk_widget_get_parent(panel->priv->table_info.table) != panel->priv->vbox_filling) {
    gtk_box_pack_start(GTK_BOX(panel->priv->vbox_filling), panel->priv->table_info.table, FALSE, TRUE, 0);
  }  
  
  for (list = panel->priv->prototypes; list; list = list->next) {
    entry = (tagEntry *)(list->data);
    hint = gedit_code_listing_plugin_panel_tag_hint(entry);
    gedit_code_listing_plugin_panel_process_tag(panel, entry, hint);
  }
  
  next = gtk_tree_model_get_iter_first(model, &iter);
  
  while (next) {
    if (!gtk_tree_model_iter_has_child(model, &iter)) {
      next = gtk_tree_store_remove(GTK_TREE_STORE(model), &iter);
    } else {
      next = gtk_tree_model_iter_next(model, &iter);
    }
  }
  
  g_list_free(panel->priv->prototypes);
  panel->priv->prototypes = NULL;
  gedit_code_listing_plugin_panel_free_entry_parents(panel);
  gtk_widget_set_sensitive(panel->priv->entry_search, TRUE);
}

/* This is actualy a GList containing tagEntries
 * The tagEntry structs contain a user_data field which is filled with a
 * GtkTreePath which should be freed
 */
static gboolean
gedit_code_listing_plugin_panel_free_entry_parent(gpointer key, gpointer value, gpointer user_data)
{
  GList *iter;
  tagEntryProxy *entry;
  
  for (iter = (GList *)(value); iter; iter = iter->next) {
    entry = (tagEntryProxy *)(iter->data);
    gtk_tree_path_free(entry->path);
    g_free(entry);
  }
  
  g_list_free((GList *)(value));
  return TRUE;
}

static void
gedit_code_listing_plugin_panel_free_entry_parents(GeditCodeListingPluginPanel *panel) {
  if (panel->priv->entry_parents != NULL) {
    g_hash_table_foreach_remove(panel->priv->entry_parents, gedit_code_listing_plugin_panel_free_entry_parent, panel);
    g_hash_table_destroy(panel->priv->entry_parents);
    panel->priv->entry_parents = NULL;
  }
  
  g_list_free(panel->priv->prototypes);
  panel->priv->prototypes = NULL;
}

static void
gedit_code_listing_plugin_panel_process_prototype(GeditCodeListingPluginPanel *panel, 
  GtkTreeStore *tstore, tagEntry *entry, enum TagHint hint, gchar *par, enum TagHint parHint, gchar *sort) {
  GtkTreeIter p, iter;
  gboolean parentFound;
  tagEntry *entryImpl;
  
  if (par != NULL) {
    parentFound = gedit_code_listing_plugin_panel_tree_find_parent(panel, GTK_TREE_MODEL(tstore), par, parHint, &p);
  } else {
    parentFound = gedit_code_listing_plugin_panel_tree_get_parent(GTK_TREE_MODEL(tstore), TAG_HINT_FUNCTION, &p);
  }
  
  if (gedit_code_listing_plugin_panel_tree_find_tag(GTK_TREE_MODEL(tstore), entry->name, TAG_HINT_FUNCTION, &p, &iter)) {
    gtk_tree_model_get(GTK_TREE_MODEL(tstore), &iter, COLUMN_TAG_ENTRY, &entryImpl, -1);
    entryImpl->prototype = entry;  
    gtk_tree_store_set(tstore, &iter, COLUMN_TAG_SORT, sort, -1);
  }
}

static void
gedit_code_listing_plugin_panel_process_tag(GeditCodeListingPluginPanel *panel,
  tagEntry *entry, enum TagHint hint) {
  
  GtkTreeStore *tstore = GTK_TREE_STORE(PANEL_GTK_TREE_MODEL(panel));
  GtkTreeIter iter, parent, p;
  gchar *par = NULL, *access = NULL, *sort = NULL, *tmp, *tmp2;
  gboolean isParent = FALSE, hasParent = FALSE;
  enum TagHint parHint = TAG_HINT_NONE;
  int ac;

  switch (hint) {
    case TAG_HINT_CLASS: case TAG_HINT_STRUCT: case TAG_HINT_ENUM:
      isParent = TRUE;
      break;
    default:
      isParent = FALSE;
      break;    
  }

  if ((par = (gchar *)(g_hash_table_lookup(entry->fields, "class")))) {
    parHint = TAG_HINT_CLASS;
  } else if ((par = (gchar *)(g_hash_table_lookup(entry->fields, "struct")))) {
    parHint = TAG_HINT_STRUCT;
  } else if ((par = (gchar *)(g_hash_table_lookup(entry->fields, "enum")))) {
    parHint = TAG_HINT_ENUM;
  }
  
  if (isParent && (tmp = (gchar *)(g_hash_table_lookup(entry->fields, "namespace")))) {
    tmp2 = g_strdup_printf("%s::%s", tmp, entry->name);
    g_free(entry->name);
    entry->name = tmp2;
  }

  if ((access = (gchar *)(g_hash_table_lookup(entry->fields, "access")))) {
    if (strcasecmp(access, "private") == 0) {
      ac = 0;
    } else if (strcasecmp(access, "protected") == 0) {
      ac = 1;
    } else {
      ac = 2;
    }
  } else {
    ac = 2;
  }
  
  if (hint == TAG_HINT_NONE) {
    sort = g_strdup_printf("%d%d%s", hint, ac, entry->name);
  } else {
    sort = g_strdup_printf("%d%d%s", ac, hint, entry->name);
  }
  
  if (hint == TAG_HINT_PROTO) {
    gedit_code_listing_plugin_panel_process_prototype(panel, tstore, entry, hint, par, parHint, sort);
  } else {
    hasParent = gedit_code_listing_plugin_panel_tree_get_parent(GTK_TREE_MODEL(tstore), hint, &parent);

	  if (par != NULL) {
	    gedit_code_listing_plugin_panel_tree_get_parent(GTK_TREE_MODEL(tstore), parHint, &parent);
	    
	    if (!gedit_code_listing_plugin_panel_tree_find_parent(panel, GTK_TREE_MODEL(tstore), par, parHint, &p)) {
	      gtk_tree_store_append(tstore, &p, &parent);
	      gedit_code_listing_plugin_panel_set_tag(panel, tstore, &p, NULL, par, parHint, sort);
	      gedit_code_listing_plugin_panel_tree_add_parent(panel, par, parHint, &p);
	    }
	    
	    gtk_tree_store_append(tstore, &iter, &p);
	    gedit_code_listing_plugin_panel_set_tag(panel, tstore, &iter, entry, entry->name, hint, sort);
	  } else if (isParent) {
	    if (gedit_code_listing_plugin_panel_tree_find_parent(panel, GTK_TREE_MODEL(tstore), entry->name, hint, &p)) {
	      gedit_code_listing_plugin_panel_set_tag(panel, tstore, &p, entry, entry->name, hint, sort);
	    } else {
	      gtk_tree_store_append(tstore, &p, &parent);
	      gedit_code_listing_plugin_panel_set_tag(panel, tstore, &p, entry, entry->name, hint, sort);
	      gedit_code_listing_plugin_panel_tree_add_parent(panel, entry->name, hint, &p);
	    }
	  } else {
	    if (hasParent) {
	      gtk_tree_store_append(tstore, &iter, &parent);
	    } else {
	      gtk_tree_store_append(tstore, &iter, NULL);
	    }
	    
	    gedit_code_listing_plugin_panel_set_tag(panel, tstore, &iter, entry, entry->name, hint, sort);
	  }
  }
  
  g_free(sort);
}

static void
gedit_code_listing_plugin_panel_tag_parsed(GeditCodeListingPluginParser *parser,
  tagEntry *entry, GeditCodeListingPluginPanel *panel) {
  enum TagHint hint;
  long position, size;
  
  gedit_code_listing_plugin_parser_parse_progress(parser, &position, &size);
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(panel->priv->hbox_progress.progress),
    (double)(position) / (double)(size));
  
  if (panel->priv->entry_parents == NULL) {
    panel->priv->entry_parents = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, NULL);
  }

  hint = gedit_code_listing_plugin_panel_tag_hint(entry);
  
  if (hint == TAG_HINT_PROTO) {
    panel->priv->prototypes = g_list_prepend(panel->priv->prototypes, entry);
  } else {
    gedit_code_listing_plugin_panel_process_tag(panel, entry, hint);
  }
}

static void
gedit_code_listing_plugin_panel_load_tag_file(GeditCodeListingPluginPanel *panel, char *path)
{
  gedit_code_listing_plugin_panel_initialize_tree(panel);
  
  if (panel->priv->parser != NULL) {
    g_object_unref(panel->priv->parser);
  }

  if (gtk_widget_get_parent(panel->priv->table_info.table) == panel->priv->vbox_filling) {
    gtk_container_remove(GTK_CONTAINER(panel->priv->vbox_filling), panel->priv->table_info.table);
  }
  if (gtk_widget_get_parent(panel->priv->hbox_progress.hbox) != panel->priv->vbox_filling) {
    gtk_box_pack_start(GTK_BOX(panel->priv->vbox_filling), panel->priv->hbox_progress.hbox, FALSE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(panel->priv->vbox_filling), panel->priv->hbox_progress.button_cancel, FALSE, TRUE, 0);
    gtk_label_set_text(GTK_LABEL(panel->priv->hbox_progress.label), "Parsing:");
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(panel->priv->hbox_progress.progress), 0.0);
  }
    
  panel->priv->parser = gedit_code_listing_plugin_parser_new(path);
  
  g_signal_connect(panel->priv->parser, "tag-parsed", (GCallback)gedit_code_listing_plugin_panel_tag_parsed, panel);
  g_signal_connect(panel->priv->parser, "parse-done", (GCallback)gedit_code_listing_plugin_panel_parse_done, panel);
  g_signal_connect(panel->priv->parser, "tag-found", (GCallback)gedit_code_listing_plugin_panel_tag_found, panel);
  g_signal_connect(panel->priv->parser, "search-done", (GCallback)gedit_code_listing_plugin_panel_search_done, panel);
  
  gtk_widget_set_sensitive(panel->priv->entry_search, FALSE);
  gedit_code_listing_plugin_parser_parse(panel->priv->parser);
}

static void
on_menu_item_activate(GtkMenuItem *widget, GeditCodeListingPluginPanel *panel)
{
  GList *children;
  int i;
  GSList *pos;
  gpointer *data;
  
  children = gtk_container_get_children(GTK_CONTAINER(panel->priv->menu));
  i = g_list_index(children, widget) - 3;
  g_list_free(children);
  
  gtk_menu_reorder_child(GTK_MENU(panel->priv->menu), GTK_WIDGET(widget), 3);
  
  /* Remove found tag and prepend it again */
  pos = g_slist_nth(panel->priv->tag_files, i);
  data = pos->data;

  panel->priv->tag_files = g_slist_delete_link(panel->priv->tag_files, pos);
  panel->priv->tag_files = g_slist_prepend(panel->priv->tag_files,data);
  
  gedit_code_listing_plugin_panel_load_tag_file(panel, g_slist_nth_data(panel->priv->tag_files, i));
}

static GtkWidget *
gedit_code_listing_plugin_panel_add_tag_file(GeditCodeListingPluginPanel *panel, char *path)
{
  GtkWidget *item;
  GList *children = gtk_container_get_children(GTK_CONTAINER(panel->priv->menu));
  GSList *pos;
  int i;
  gchar *basename;
  
  pos = g_slist_find_custom(panel->priv->tag_files, path, (GCompareFunc)strcmp);
  
  if (pos) {
    /* Move menu item up */
    item = g_list_nth_data(children, g_slist_position(panel->priv->tag_files, pos) + 3);
  } else {
    /* First remove leftover tags/menuitems */
	  for (i = 4; i < g_slist_length(panel->priv->tag_files); i++) {
	    panel->priv->tag_files = g_slist_delete_link(panel->priv->tag_files, g_slist_last(panel->priv->tag_files));
	    gtk_widget_destroy(g_list_nth_data(children, i + 3));
	  }
	  
	  /* Now add the new tag file */
	  basename = g_path_get_basename(path);
	  item = gtk_menu_item_new_with_label(basename);
	  g_free(basename);

	  gtk_menu_shell_insert(GTK_MENU_SHELL(panel->priv->menu), item, 3);
    gtk_widget_show(item);
    
	  panel->priv->tag_files = g_slist_prepend(panel->priv->tag_files, g_strdup(path));
    g_signal_connect(item, "activate", G_CALLBACK(on_menu_item_activate), panel);
	}
	
	g_list_free(children);
	return item;
}

static void
on_menu_item_open_activate(GtkMenuItem *widget, GeditCodeListingPluginPanel *panel)
{
  char *path;
  GtkWidget *d, *item;
  
  if (panel->priv->prev_path == NULL) {
    panel->priv->prev_path = gconf_client_get_string(gedit_prefs_manager->gconf_client,
      GPM_PLUGIN_CODE_LISTING "/prev_path", NULL);
  }
  
  /*  Filechooser dialog */
  d = gtk_file_chooser_dialog_new(_("Open tags file"), 
    GTK_WINDOW(panel->priv->window), GTK_FILE_CHOOSER_ACTION_OPEN, 
    GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
    GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
  gtk_file_chooser_set_local_only(GTK_FILE_CHOOSER(d), TRUE);
  
  if (panel->priv->prev_path != NULL) {
    gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(d), panel->priv->prev_path);
  }
  
  if (gtk_dialog_run(GTK_DIALOG(d)) == GTK_RESPONSE_ACCEPT) {
    path = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(d));
  } else {
    path = NULL;
  }
  
  gtk_widget_destroy(d);
  
  if (path != NULL) {
    item = gedit_code_listing_plugin_panel_add_tag_file(panel, path);
    gtk_menu_item_activate(GTK_MENU_ITEM(item));
  }
  
  g_free(path);
}

static void
on_menu_item_refresh_activate(GtkMenuItem *widget, GeditCodeListingPluginPanel *panel)
{

}

static void
gedit_code_listing_plugin_panel_populate_tag_files(GeditCodeListingPluginPanel *panel)
{
  GSList *tagfiles, *iter;
  GList *children;
  
  tagfiles = gconf_client_get_list(gedit_prefs_manager->gconf_client, 
      GPM_PLUGIN_CODE_LISTING "/tag_files", GCONF_VALUE_STRING, NULL);

  tagfiles = g_slist_reverse(tagfiles);
  
  for (iter = tagfiles; iter != NULL; iter = iter->next) {
    gedit_code_listing_plugin_panel_add_tag_file(panel, iter->data);
  }
  
  if (g_slist_length(panel->priv->tag_files) > 0) {
    children = gtk_container_get_children(GTK_CONTAINER(panel->priv->menu));
    if (g_list_length(children) > 3) {
      gtk_menu_item_activate(GTK_MENU_ITEM(g_list_nth_data(children, 3)));
    }
    g_list_free(children);
  }
  
  gedit_code_listing_plugin_panel_gslist_free(tagfiles);
}

static GtkWidget *
gedit_code_listing_plugin_panel_create_menu(GeditCodeListingPluginPanel *panel) 
{
  GtkWidget *menubar;
  GtkWidget *item;
  
  panel->priv->menu = gtk_menu_new();
  menubar = gtk_menu_bar_new();
  item = gtk_menu_item_new_with_label(_("Tag file"));
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), GTK_WIDGET(panel->priv->menu));
  gtk_container_add(GTK_CONTAINER(menubar), item);
  
  item = gtk_image_menu_item_new_from_stock(GTK_STOCK_OPEN, NULL);
  gtk_menu_shell_append(GTK_MENU_SHELL(panel->priv->menu), item);
  g_signal_connect(item, "activate", G_CALLBACK(on_menu_item_open_activate), panel);
  
  item = gtk_image_menu_item_new_from_stock(GTK_STOCK_REFRESH, NULL);
  gtk_menu_shell_append(GTK_MENU_SHELL(panel->priv->menu), item);
  g_signal_connect(item, "activate", G_CALLBACK(on_menu_item_refresh_activate), panel);
  
  item = gtk_separator_menu_item_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(panel->priv->menu), item);
  
  return menubar;
}

static void
gedit_code_listing_plugin_panel_load_pixbufs(GeditCodeListingPluginPanel *panel)
{
  int i;
  
  for (i = 0; i < TAG_HINT_COUNT; i++) {
    panel->priv->pixbufs[i] = gdk_pixbuf_new_from_file(GeditListingTypes[i].pixbuf, NULL);
  }
}

static GeditDocument *
gedit_code_listing_plugin_panel_find_document(GeditCodeListingPluginPanel *panel, gchar *uri) {
  GList *documents, *doc;
  GeditDocument *gdoc = NULL;
  gchar *doc_uri;
  
  documents = gedit_window_get_documents(panel->priv->window);
  
  for (doc = documents; doc; doc = doc->next) {
    gdoc = (GeditDocument *)(doc->data);
    doc_uri = gedit_document_get_uri(gdoc); 
    
    if (doc_uri != NULL && gnome_vfs_uris_match(doc_uri, uri)) {
      g_free(doc_uri);
      break;
    } else {
      g_free(doc_uri);
      gdoc = NULL;
    }
  }
  
  g_list_free(documents);
  return gdoc;
}

static void
gedit_code_listing_plugin_panel_goto_entry(GeditCodeListingPluginPanel *panel, tagEntry *entry) {
  gchar *base_uri, *all_uri, *base, *base_with, *pattern;
  GeditDocument *gdoc;
  GeditTab *tab;
  GSList *uris;
  GtkTextIter start, end, sMatch, eMatch;
  GtkTextBuffer *buffer;
  
  /* Construct uri */
  base = gedit_code_listing_plugin_parser_dirname(panel->priv->parser);
  base_with = g_strconcat(base, G_DIR_SEPARATOR_S, NULL);
 
  base_uri = g_filename_to_uri(base_with, NULL, NULL);
  all_uri = gnome_vfs_uri_make_full_from_relative(base_uri, entry->file);
  
  gdoc = gedit_code_listing_plugin_panel_find_document(panel, all_uri);
  
  if (gdoc == NULL) {
    uris = g_slist_prepend(NULL, all_uri);
    gedit_cmd_load_files(panel->priv->window, uris, NULL);
    gdoc = gedit_code_listing_plugin_panel_find_document(panel, all_uri);
    g_slist_free(uris);
  }
  
  if (gdoc != NULL) {
    tab = gedit_tab_get_from_document(gdoc);
    gedit_window_set_active_tab(panel->priv->window, tab);

    while (g_main_context_pending(NULL)) {
      g_main_context_iteration(NULL, FALSE);
    }

    if (entry->address.lineNumber > 0) {
      gedit_document_goto_line(gdoc, entry->address.lineNumber - 1);
    } else {
      buffer = GTK_TEXT_BUFFER(gdoc);
      gtk_text_buffer_get_start_iter(buffer, &start);
      gtk_text_buffer_get_end_iter(buffer, &end);
      
      pattern = g_strndup(entry->address.pattern + 2, strlen(entry->address.pattern) - 4);
      gedit_document_set_search_text(gdoc, pattern, GEDIT_SEARCH_CASE_SENSITIVE);
      gedit_document_goto_line(gdoc, 0);
      
      if (gedit_document_search_forward(gdoc, &start, &end, &sMatch, &eMatch)) {
        gtk_text_buffer_place_cursor(buffer, &sMatch);
        gtk_text_buffer_move_mark_by_name(buffer, "selection_bound", &eMatch);
        gedit_view_scroll_to_cursor(gedit_window_get_active_view(panel->priv->window));
      } else {

      }
      
      g_free(pattern);
    }
  } else {
    /* File could not be found or something */
  }
 
  g_free(base_with);
  g_free(all_uri);
  g_free(base_uri);
  g_free(base);
}

static void
on_tree_view_tags_row_activated(GtkTreeView *treeview, GtkTreePath *arg1, GtkTreeViewColumn *arg2, GeditCodeListingPluginPanel *panel) {
  GtkTreeIter iter;
  GtkTreeModel *model;
  tagEntry *entry;
  GdkModifierType mask;
  
  model = gtk_tree_view_get_model(GTK_TREE_VIEW(panel->priv->tree_view_tags));
  gdk_device_get_state(gdk_device_get_core_pointer(), GTK_WIDGET(panel->priv->window)->window, NULL, &mask);
  
  if (model == panel->priv->tree_store_full) {
    gtk_tree_model_get_iter(model, &iter, arg1);
    gtk_tree_model_get(model, &iter, COLUMN_TAG_ENTRY, &entry, -1);
  } else {
    gtk_tree_model_get_iter(model, &iter, arg1);
    gtk_tree_model_get(model, &iter, SEARCH_COLUMN_TAG_ENTRY, &entry, -1);
  }
  
  if (entry != NULL) {
    if (mask & GDK_CONTROL_MASK && entry->prototype != NULL) {
      gedit_code_listing_plugin_panel_goto_entry(panel, entry->prototype);
    } else {
      gedit_code_listing_plugin_panel_goto_entry(panel, entry);
    }
  }
}

static void
on_tree_view_tags_row_changed(GtkTreeSelection *treeselection, GeditCodeListingPluginPanel *panel) {
  GtkTreeIter iter;
  GtkTreeModel *model;
  tagEntry *entry;
  gchar *line;
  
  if (gtk_tree_selection_get_selected(treeselection, &model, &iter)) {
    if (gtk_tree_view_get_model(GTK_TREE_VIEW(panel->priv->tree_view_tags)) == panel->priv->tree_store_full) {
      gtk_tree_model_get(model, &iter, COLUMN_TAG_ENTRY, &entry, -1);
    } else {
      gtk_tree_model_get(model, &iter, SEARCH_COLUMN_TAG_ENTRY, &entry, -1);
    }
  
	  if (entry != NULL) {
	    gtk_label_set_text(GTK_LABEL(panel->priv->table_info.label_file), entry->file);
	    
	    if (entry->address.lineNumber > 0) {
  	    line = g_strdup_printf("%ld", entry->address.lineNumber);
  	  } else {
  	    line = g_strdup("<i>unknown</i>");
  	  }
  	  
	    gtk_label_set_markup(GTK_LABEL(panel->priv->table_info.label_line), line);
	    g_free(line);
	  } else {
      gtk_label_set_markup(GTK_LABEL(panel->priv->table_info.label_file), "");
	    gtk_label_set_markup(GTK_LABEL(panel->priv->table_info.label_line), "");
	  }
	}
}

static void
on_entry_search_activate(GtkEntry *entry, GeditCodeListingPluginPanel *panel)
{
  const gchar *query = gtk_entry_get_text(entry);
  
  if (strlen(query) == 0) {
    /* Show full tree */
    if (gtk_tree_view_get_model(GTK_TREE_VIEW(panel->priv->tree_view_tags)) != panel->priv->tree_store_full) {
      gtk_tree_view_set_model(GTK_TREE_VIEW(panel->priv->tree_view_tags), NULL);
      gtk_tree_view_remove_column(GTK_TREE_VIEW(panel->priv->tree_view_tags), panel->priv->column_search);
      gtk_tree_view_append_column(GTK_TREE_VIEW(panel->priv->tree_view_tags), panel->priv->column_full);
      gtk_tree_view_set_search_column(GTK_TREE_VIEW (panel->priv->tree_view_tags), COLUMN_TAG_SORT);
      gtk_tree_view_set_model(GTK_TREE_VIEW(panel->priv->tree_view_tags), panel->priv->tree_store_full);
    }
  } else {
    if (gtk_widget_get_parent(panel->priv->table_info.table) == panel->priv->vbox_filling) {
      gtk_container_remove(GTK_CONTAINER(panel->priv->vbox_filling), panel->priv->table_info.table);
    }
    if (gtk_widget_get_parent(panel->priv->hbox_progress.hbox) != panel->priv->vbox_filling) {
      gtk_box_pack_start(GTK_BOX(panel->priv->vbox_filling), panel->priv->hbox_progress.hbox, FALSE, TRUE, 0);
      gtk_box_pack_start(GTK_BOX(panel->priv->vbox_filling), panel->priv->hbox_progress.button_cancel, FALSE, TRUE, 0);
      gtk_label_set_text(GTK_LABEL(panel->priv->hbox_progress.label), _("Search:"));
      gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(panel->priv->hbox_progress.progress), 0.0);
    }
	  
    gtk_list_store_clear(GTK_LIST_STORE(PANEL_GTK_TREE_MODEL_SEARCH(panel)));
   
    if (gtk_tree_view_get_model(GTK_TREE_VIEW(panel->priv->tree_view_tags)) != panel->priv->list_store_search) {
      gtk_tree_view_set_model(GTK_TREE_VIEW(panel->priv->tree_view_tags), NULL);
      gtk_tree_view_remove_column(GTK_TREE_VIEW(panel->priv->tree_view_tags), panel->priv->column_full);
      gtk_tree_view_append_column(GTK_TREE_VIEW(panel->priv->tree_view_tags), panel->priv->column_search);
      gtk_tree_view_set_search_column(GTK_TREE_VIEW(panel->priv->tree_view_tags), SEARCH_COLUMN_TAG_NAME);
      gtk_tree_view_set_model(GTK_TREE_VIEW(panel->priv->tree_view_tags), panel->priv->list_store_search);
    }

    gtk_widget_set_sensitive(GTK_WIDGET(entry), FALSE);
    gedit_code_listing_plugin_parser_search(panel->priv->parser, (gchar *)(query));
  }
}

static void
on_button_cancel_clicked(GtkButton *button, GeditCodeListingPluginPanel *panel) {
  gedit_code_listing_plugin_parser_cancel(panel->priv->parser);
}

static void
gedit_code_listing_plugin_panel_create_ui(GeditCodeListingPluginPanel *panel)
{	
  GtkWidget *box, *vbox, *item, *sw, *menu;
  GtkTreeViewColumn *column;
  GtkCellRenderer *cell;
  GList *focus_chain = NULL;
  GtkTreeModel *store;
  GtkTreeIter iter;
  
  /* Create menu */
  menu = gedit_code_listing_plugin_panel_create_menu(panel);
  gtk_box_pack_start(GTK_BOX(panel), menu, FALSE, TRUE, 0);

  /* Create the container bin */
  vbox = gtk_vbox_new(FALSE, 6);
  panel->priv->vbox_filling = vbox;
  
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 6);
  gtk_box_pack_start(GTK_BOX(panel), vbox, TRUE, TRUE, 0);
  
  /* Create the search box */
  box = gtk_hbox_new(FALSE, 6);
  gtk_box_pack_start(GTK_BOX(vbox), box, FALSE, TRUE, 0);
  
  item = gtk_label_new(_("Search:"));
  gtk_misc_set_alignment(GTK_MISC(item), 0.0, 0.5);
  gtk_box_pack_start(GTK_BOX(box), item, FALSE, TRUE, 0);
  
  panel->priv->entry_search = gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(box), panel->priv->entry_search, TRUE, TRUE, 0);

  gtk_tooltips_set_tip (panel->priv->tooltips, panel->priv->entry_search,
      _("Enter the class/function/struct to search for and press `enter'. Leave empty and press `enter' to go back to the full tree"),
			      NULL); 
			      
  /* Create the listview */
  sw = gtk_scrolled_window_new (NULL, NULL);
	
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
      GTK_POLICY_AUTOMATIC,
      GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw),
                                             GTK_SHADOW_IN);
  gtk_box_pack_start (GTK_BOX (vbox), sw, TRUE, TRUE, 0);

  /* Create tree store */
  store = GTK_TREE_MODEL(gtk_tree_store_new(NUM_COLUMNS, GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_POINTER, G_TYPE_INT, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT));
  panel->priv->tree_store_full = gtk_tree_model_sort_new_with_model(store);
  g_object_ref(panel->priv->tree_store_full);
  
  store = GTK_TREE_MODEL(gtk_list_store_new(SEARCH_NUM_COLUMNS, GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_POINTER));
  panel->priv->list_store_search = gtk_tree_model_sort_new_with_model(store);
  g_object_ref(panel->priv->list_store_search);

  /* Create treeview */
  panel->priv->tree_view_tags = gtk_tree_view_new_with_model(panel->priv->tree_store_full);  
  gtk_container_add(GTK_CONTAINER (sw), panel->priv->tree_view_tags);
  
  /* Setup treeview */
  gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (panel->priv->tree_view_tags), FALSE);
  gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (panel->priv->tree_view_tags), FALSE);
  gtk_tree_view_set_enable_search(GTK_TREE_VIEW(panel->priv->tree_view_tags), TRUE);
  gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(panel->priv->tree_store_full),
                                        COLUMN_TAG_SORT, GTK_SORT_ASCENDING);
  gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(panel->priv->list_store_search),
                                        SEARCH_COLUMN_TAG_NAME, GTK_SORT_ASCENDING);
  
  gtk_tooltips_set_tip (panel->priv->tooltips, panel->priv->tree_view_tags,
      _("Double-click on a tag to goto the implementation, or Ctrl-double-click to go to its definition (if appropriate)"), NULL);
			      
  /* Add the tags column */
  cell = gtk_cell_renderer_pixbuf_new();
  column = gtk_tree_view_column_new_with_attributes(NULL, cell, "pixbuf",
      COLUMN_TAG_PIXBUF, NULL);
  gtk_tree_view_append_column(GTK_TREE_VIEW(panel->priv->tree_view_tags),
      column);
                 
  cell = gtk_cell_renderer_text_new ();
  panel->priv->column_full = gtk_tree_view_column_new_with_attributes (_("Tags"), 
      cell, "text", COLUMN_TAG_NAME, "style", COLUMN_TAG_STYLE, "weight",
      COLUMN_TAG_WEIGHT, NULL);
  cell = gtk_cell_renderer_text_new();
  panel->priv->column_search = gtk_tree_view_column_new_with_attributes(_("Tags"),
     cell,"text", SEARCH_COLUMN_TAG_NAME, NULL);

  g_object_ref(panel->priv->column_full);
  g_object_ref(panel->priv->column_search);
  
  gtk_tree_view_append_column (GTK_TREE_VIEW (panel->priv->tree_view_tags), 
      panel->priv->column_full);
  gtk_tree_view_set_search_column (GTK_TREE_VIEW (panel->priv->tree_view_tags),
    COLUMN_TAG_SORT);
  
  if (g_slist_length(panel->priv->tag_files) == 0) {
    store = gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(panel->priv->tree_store_full));
    gtk_tree_store_prepend(GTK_TREE_STORE(store), &iter, NULL);
    gtk_tree_store_set(GTK_TREE_STORE(store), &iter, COLUMN_TAG_NAME, _("Select a tag file above"), COLUMN_TAG_STYLE, PANGO_STYLE_ITALIC, COLUMN_TAG_ENTRY, NULL, -1);
  }
  
  panel->priv->table_info.table = gtk_table_new(2, 2, FALSE);
  gtk_table_set_row_spacings(GTK_TABLE(panel->priv->table_info.table), 3);
  gtk_table_set_col_spacings(GTK_TABLE(panel->priv->table_info.table), 6);
  
  item = gtk_label_new(NULL);
  gtk_misc_set_alignment(GTK_MISC(item), 0.0, 0.5);
  gtk_label_set_markup(GTK_LABEL(item), _("<b>File:</b>"));
  gtk_table_attach(GTK_TABLE(panel->priv->table_info.table), item, 0, 1, 0, 1, GTK_FILL | GTK_SHRINK, GTK_FILL | GTK_SHRINK, 0, 0);
  
  panel->priv->table_info.label_file = gtk_label_new(NULL);
  gtk_misc_set_alignment(GTK_MISC(panel->priv->table_info.label_file), 0.0, 0.5);
  gtk_table_attach(GTK_TABLE(panel->priv->table_info.table), panel->priv->table_info.label_file, 1, 2, 0, 1, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_SHRINK, 0, 0);
  
  item = gtk_label_new(NULL);
  gtk_misc_set_alignment(GTK_MISC(item), 0.0, 0.5);
  gtk_label_set_markup(GTK_LABEL(item), _("<b>Line:</b>"));
  gtk_table_attach(GTK_TABLE(panel->priv->table_info.table), item, 0, 1, 1, 2, GTK_FILL | GTK_SHRINK, GTK_FILL | GTK_SHRINK, 0, 0);
  
  panel->priv->table_info.label_line = gtk_label_new(NULL);
  gtk_misc_set_alignment(GTK_MISC(panel->priv->table_info.label_line), 0.0, 0.5);
  gtk_table_attach(GTK_TABLE(panel->priv->table_info.table), panel->priv->table_info.label_line, 1, 2, 1, 2, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_SHRINK, 0, 0); 
  
  g_object_ref(panel->priv->table_info.table);
  gtk_widget_show_all(panel->priv->table_info.table);
	
  panel->priv->hbox_progress.hbox = gtk_hbox_new(FALSE, 6);
  panel->priv->hbox_progress.label = gtk_label_new(NULL);
  gtk_box_pack_start(GTK_BOX(panel->priv->hbox_progress.hbox), panel->priv->hbox_progress.label, FALSE, TRUE, 0);
	
  panel->priv->hbox_progress.progress = gtk_progress_bar_new();
  gtk_box_pack_start(GTK_BOX(panel->priv->hbox_progress.hbox), panel->priv->hbox_progress.progress, TRUE, TRUE, 0);

  g_object_ref(panel->priv->hbox_progress.hbox);	
  gtk_widget_show_all(panel->priv->hbox_progress.hbox);
	
  panel->priv->hbox_progress.button_cancel = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
  gtk_widget_show(panel->priv->hbox_progress.button_cancel);
  g_object_ref(panel->priv->hbox_progress.button_cancel);

  gtk_tooltips_set_tip (panel->priv->tooltips,
      panel->priv->hbox_progress.button_cancel,
      _("Cancel the current parse or search action"),
      NULL);
			      			  
  /* Set ATK descriptions/relations */
  gedit_utils_set_atk_name_description(panel->priv->entry_search, 
      _("Search field"), NULL);
  gedit_utils_set_atk_name_description(panel->priv->tree_view_tags, 
      _("Tags listing"), NULL);
  gedit_utils_set_atk_name_description(panel->priv->hbox_progress.button_cancel, 
      _("Cancel button"), NULL);
  gedit_utils_set_atk_relation(panel->priv->entry_search, 
      panel->priv->tree_view_tags, ATK_RELATION_CONTROLLER_FOR);
  gedit_utils_set_atk_relation(panel->priv->tree_view_tags, 
      panel->priv->entry_search, ATK_RELATION_CONTROLLED_BY);	

  focus_chain = g_list_prepend(focus_chain, panel->priv->tree_view_tags);
  focus_chain = g_list_prepend(focus_chain, panel->priv->entry_search);
	
  gtk_container_set_focus_chain(GTK_CONTAINER (panel), focus_chain);
  g_list_free(focus_chain);

  gedit_code_listing_plugin_panel_populate_tag_files(panel);
 
  /* Setup signals */
  g_signal_connect(panel->priv->tree_view_tags, "row-activated",
    G_CALLBACK(on_tree_view_tags_row_activated), panel);
  g_signal_connect(gtk_tree_view_get_selection(GTK_TREE_VIEW(panel->priv->tree_view_tags)), "changed",
    G_CALLBACK(on_tree_view_tags_row_changed), panel);
  g_signal_connect(panel->priv->entry_search, "activate",
    G_CALLBACK(on_entry_search_activate), panel);
  g_signal_connect(panel->priv->hbox_progress.button_cancel, "clicked",
    G_CALLBACK(on_button_cancel_clicked), panel);
	gtk_widget_show_all (GTK_WIDGET (panel));
}

static void
gedit_code_listing_plugin_panel_init (GeditCodeListingPluginPanel *panel)
{	
   gedit_debug (DEBUG_PLUGINS);

  panel->priv = GEDIT_CODE_LISTING_PLUGIN_PANEL_GET_PRIVATE (panel);
  panel->priv->tag_files = NULL;
  panel->priv->prev_path = NULL;
  panel->priv->parser = NULL;
  panel->priv->entry_parents = NULL;
  panel->priv->prototypes = NULL;
  
  panel->priv->tooltips = gtk_tooltips_new ();
  g_object_ref (G_OBJECT (panel->priv->tooltips));
  gtk_object_sink (GTK_OBJECT (panel->priv->tooltips));

  /* Load pixbufs */
  gedit_code_listing_plugin_panel_load_pixbufs(panel);
  
  /* Build the window content */
  gedit_code_listing_plugin_panel_create_ui(panel);	
}

GtkWidget *
gedit_code_listing_plugin_panel_new (GeditWindow *window)
{
  GtkWidget *result;
  g_return_val_if_fail (GEDIT_IS_WINDOW (window), NULL);

  result = GTK_WIDGET (g_object_new (GEDIT_TYPE_CODE_LISTING_PLUGIN_PANEL, 
      "window", window, NULL));
					 
  gtk_box_set_spacing(GTK_BOX(result), 0);
  gtk_container_set_border_width(GTK_CONTAINER(result), 0);
  return result;
}

GType
gedit_code_listing_plugin_panel_get_type (void)
{
  return type;
}

GType
gedit_code_listing_plugin_panel_register_type (GTypeModule *module)
{
  static const GTypeInfo our_info =
  {
    sizeof (GeditCodeListingPluginPanelClass),
    NULL, /* base_init */
    NULL, /* base_finalize */
    (GClassInitFunc) gedit_code_listing_plugin_panel_class_init,
    NULL,
    NULL, /* class_data */
    sizeof (GeditCodeListingPluginPanel),
    0, /* n_preallocs */
    (GInstanceInitFunc) gedit_code_listing_plugin_panel_init
  };

    gedit_debug_message (DEBUG_PLUGINS, "Registering GeditCodeListingPluginPanel");

    type = g_type_module_register_type (module, GTK_TYPE_VBOX,
        "GeditCodeListingPluginPanel", &our_info, 0);
	return type;
}
