/*
*   $Id: readtags.h,v 1.12 2003/04/01 05:02:21 darren Exp $
*
*   Copyright (c) 1996-2003, Darren Hiebert
*
*   This source code is released for the public domain.
*
*   This file defines the public interface for looking up tag entries in tag
*   files.
*
*   The functions defined in this interface are intended to provide tag file
*   support to a software tool. The tag lookups provided are sufficiently fast
*   enough to permit opening a sorted tag file, searching for a matching tag,
*   then closing the tag file each time a tag is looked up (search times are
*   on the order of hundreths of a second, even for huge tag files). This is
*   the recommended use of this library for most tool applications. Adhering
*   to this approach permits a user to regenerate a tag file at will without
*   the tool needing to detect and resynchronize with changes to the tag file.
*   Even for an unsorted 24MB tag file, tag searches take about one second.
* 
*   ### I'm not sure how to license this (because I used some of Darren's code
*   ### but almost everything is a rewrite and mostly I just looked at his code
*   ### Some advice needed, Jesse van den Kieboom
*/
#ifndef __GEDIT_CODE_LISTING_PLUGIN_PARSER_H__
#define __GEDIT_CODE_LISTING_PLUGIN_PARSER_H__

#include <glib.h>
#include <glib-object.h>

#ifdef __cplusplus
extern "C" {
#endif

G_BEGIN_DECLS

/*
 * Type checking and casting macros
 */
#define GEDIT_TYPE_CODE_LISTING_PLUGIN_PARSER              (gedit_code_listing_plugin_parser_get_type())
#define GEDIT_CODE_LISTING_PLUGIN_PARSER(obj)              (G_TYPE_CHECK_INSTANCE_CAST((obj), GEDIT_TYPE_CODE_LISTING_PLUGIN_PARSER, GeditCodeListingPluginParser))
#define GEDIT_CODE_LISTING_PLUGIN_PARSER_CONST(obj)        (G_TYPE_CHECK_INSTANCE_CAST((obj), GEDIT_TYPE_CODE_LISTING_PLUGIN_PARSER, GeditCodeListingPluginParser const))
#define GEDIT_CODE_LISTING_PLUGIN_PARSER_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST((klass), GEDIT_TYPE_CODE_LISTING_PLUGIN_PARSER, GeditCodeListingPluginParserClass))
#define GEDIT_IS_CODE_LISTING_PLUGIN_PARSER(obj)           (G_TYPE_CHECK_INSTANCE_TYPE((obj), GEDIT_TYPE_CODE_LISTING_PLUGIN_PARSER))
#define GEDIT_IS_CODE_LISTING_PLUGIN_PARSER_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), GEDIT_TYPE_CODE_LISTING_PLUGIN_PARSER))
#define GEDIT_CODE_LISTING_PLUGIN_PARSER_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS((obj), GEDIT_TYPE_CODE_LISTING_PLUGIN_PARSER, GeditCodeListingPluginParserClass))

/* Private structure type */
typedef struct _GeditCodeListingPluginParserPrivate GeditCodeListingPluginParserPrivate;

/*
 * Main object structure
 */
typedef struct _GeditCodeListingPluginParser GeditCodeListingPluginParser;

struct _GeditCodeListingPluginParser 
{
	GObject parent_instance;

	/*< private > */
	GeditCodeListingPluginParserPrivate *priv;
};

/*
 * Class definition
 */
typedef struct _GeditCodeListingPluginParserClass GeditCodeListingPluginParserClass;
typedef struct _tagEntry tagEntry;
struct _GeditCodeListingPluginParserClass 
{
	GObjectClass parent_class;
	
	/* Signals */
	void (*tag_parsed)  (GeditCodeListingPluginParser *parser, 
			      const tagEntry *entry); 
	void (*parse_done)  (GeditCodeListingPluginParser *parser);
	void (*tag_found)  (GeditCodeListingPluginParser *parser, 
			      const tagEntry *entry);
	void (*search_done)  (GeditCodeListingPluginParser *parser);
};

/*
 * Public methods
 */
GType gedit_code_listing_plugin_parser_register_type(GTypeModule *module);
GType gedit_code_listing_plugin_parser_get_type(void);

G_END_DECLS

/* Options for sorting */
typedef enum {
    TAG_UNSORTED, TAG_SORTED, TAG_FOLDSORTED
} sortType ;

/* Information about current tag file */
typedef struct _GeditCodeListingPluginParserTagFile {
	short format; /**< format of tag file */
	sortType sortMethod; /**< how is the tag file sorted? */
	FILE* fd; /**< pointer to file structure */
	long size; /**< the tags file size */
	long position; /** the current position in the tags file */
	gchar *path; /**< tag file path */
	GList *tags; /**< All tags */
	long ntags; /**< Length of tags */

	/* buffers to be freed at close */
	struct {
		char *author; /**< name of program author */
		char *name; /**< name of program */
		char *url; /**< URL of distribution */
		char *version; /** program version */
	} program;
} GeditCodeListingPluginParserTagFile;

/** This structure contains information about a specific tag. */
struct _tagEntry {
	gchar *name; /**< name of tag */
	gchar *file; /**< path of source file containing definition of tag */

	/** address for locating tag in source file */
	struct {
		char *pattern; /**< pattern for locating source line (may be NULL if not present) */
		unsigned long lineNumber; /**< line number in source file of tag definition (may be zero if not known) */
	} address;
	
	/** prototype holder, which can be set to point to the prototype entry */
	tagEntry *prototype;

	gchar *kind; /**< kind of tag (may by name, character, or NULL if not known) */
	short fileScope; /**< is tag of file-limited scope? */

  GHashTable *fields; /**< contains tag extension fields */
};

/** Creates a new GeditCodeListingPluginParser.
  * @param path the tags file to parse
  * @return a new GeditCodeListingPluginParser object
  */
GeditCodeListingPluginParser *gedit_code_listing_plugin_parser_new(gchar *path);

/** Get all tags.
  * Returns a list with all tags.
  * @param parser a GeditCodeListingPluginParser
  * @return a GList with all the tags (list must be freed with g_list_free)
  */
GList *gedit_code_listing_plugin_parser_tags(GeditCodeListingPluginParser *parser);

/** Parses the tags file.
  * Parses the tags file given at ..._new(gchar *)
  * @param parser a GeditCodeListingPluginParser
  * @return TRUE if parsing succeeded, otherwise FALSE is returned and error is set
  */
gboolean gedit_code_listing_plugin_parser_parse(GeditCodeListingPluginParser *parser);

/** Find tags by a certain key.
  * Finds tags by a certain key and returns the list of found tags.
  * @param parser a GeditCodeListingPluginParser
  * @param key the tag to search for
  * @return a GList containing the found tags, or NULL if no tags found (list must be freed with g_list_free)
  */
void gedit_code_listing_plugin_parser_search(GeditCodeListingPluginParser *parser, gchar *key);

/** Cancel current parse or search.
  * Cancels the current parse or search process.
  * @param parser a GeditCodeListingPluginParser
  */
void gedit_code_listing_plugin_parser_cancel(GeditCodeListingPluginParser *parser);

/** Get parse progress.
  * Gets the parse process progress. The number of bytes parsed and to total 
  * total number of bytes to parse.
  * @param parser a GeditCodeListingPluginParser
  * @param position a pointer to store the bytes parsed
  * @param size a pointer to store the total number of bytes
  */
void gedit_code_listing_plugin_parser_parse_progress(GeditCodeListingPluginParser *parser, long *position, long *size);

/** Get search progress.
  * Gets the search process progress. The number of entries searched and the 
  * total number of entries to search.
  * @param parser a GeditCodeListingPluginParser
  * @param position a pointer to store the entries searched
  * @param size a pointer to store the total number of entries
  */
void gedit_code_listing_plugin_parser_search_progress(GeditCodeListingPluginParser *parser, long *position, long *size);

/** Get the dirname where the tagfile is stored.
  * Gets the dirname where the tagfile is stored.
  * @param parser a GeditCodeListingPluginParser
  * @return a newly allocated gchar containing the dirname
  */
gchar *gedit_code_listing_plugin_parser_dirname(GeditCodeListingPluginParser *parser);

#ifdef __cplusplus
};
#endif

#endif

/* vi:set tabstop=8 shiftwidth=4: */
